/***
 * Juliac
 * Copyright (C) 2010-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.mind;

import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.type.BasicComponentType;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.api.ADLParserSupportItf;
import org.objectweb.fractal.juliac.api.ClassGeneratorItf;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.AttributeDesc;
import org.objectweb.fractal.juliac.desc.BindingDesc;
import org.objectweb.fractal.juliac.desc.BindingType;
import org.objectweb.fractal.juliac.desc.ComponentDesc;
import org.objectweb.fractal.juliac.desc.ComponentDescFactoryClassGenerator;
import org.ow2.mind.adl.graph.ComponentGraph;

/**
 * Parser for the MIND Architecture Description Language.
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2.4
 */
public class MindADLParserSupportImpl implements ADLParserSupportItf {

    // -----------------------------------------------------------------------
    // Implementation of the JuliacModuleItf interface
    // -----------------------------------------------------------------------

	final public static Class<ADLParserSupportItf> SERVICE_TYPE =
		ADLParserSupportItf.class;
	
	private Juliac jc;
	
    /**
     * Initialize the service.
      */
    public void init( Juliac jc ) {
    	this.jc = jc;
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.register(SERVICE_TYPE,this);
    }
    
    public void close( Juliac jc ) {
        JuliacConfig jconf = jc.getJuliacConfig();
        jconf.unregister(SERVICE_TYPE,this);
    }

    /**
     * Return <code>true</code> if the current source code generator handles the
     * specified ADL descriptor.
     * 
     * @param adl  the ADL descriptor to be checked
     * @return     if the ADL descriptor should be accepted
     */
    public boolean acceptADLDesc( String adl ) {
        /*
         * Don't use File#separatorChar (even on Windows) as it leads to is
         * being null when the resource is stored in a jar file.
         */
        String adlFileName = adl.replace('.','/')+".adl";
        JuliacConfig jconf = jc.getJuliacConfig();
        ClassLoader loader = jconf.getClassLoader();
        InputStream is = loader.getResourceAsStream(adlFileName);
        boolean exist = (is != null);
        return exist;
    }
    
    // -----------------------------------------------------------------------
    // Implementation of the ADLParserSupportItf interface
    // -----------------------------------------------------------------------

    /**
     * Parse the specified ADL descriptor and return a model of it.
     * 
     * @param adl      the fully-qualified name of the ADL descriptor
     * @param context  contextual information
     * @return         the model of the specified ADL
     */
    public ComponentDesc parse( String adl, Map<Object,Object> context )
	throws IOException {
        
    	try {
            ComponentDesc cdesc = toComponentDesc(adl);
            return cdesc;
    	}
    	catch( Exception e ) {
    		if( e instanceof IOException ) {
    			throw (IOException) e;
    		}
    		if( e instanceof RuntimeException ) {
    			throw (RuntimeException) e;
    		}
    		IOException ioe = new IOException();
    		ioe.initCause(e);
    		throw ioe;
    	}
    }

    /**
     * Generate the source code of a factory class for the specified ADL
     * descriptor.
     * 
     * @param adl         the fully-qualified name of the ADL descriptor
     * @param targetname  the fully-qualified name of the factory class
     */
    public void generate( String adl, String targetname )
    throws IOException {
    	
    	ComponentDesc cdesc = parse(adl,null);
        generate(cdesc,targetname);
    }
    
    /**
     * Generate the source code of a factory class for the specified model of
     * the ADL.
     * 
     * @param cdesc       the model of the ADL descriptor
     * @param targetname  the fully-qualified name of the factory class
     * @since 2.5
     */
    public void generate( ComponentDesc cdesc, String targetname )
    throws IOException {
    	
        ClassGeneratorItf cg =
    		new ComponentDescFactoryClassGenerator(jc,cdesc,targetname);
        jc.generateSourceCode(cg);
    }

    
    // ----------------------------------------------------------------------
    // Methods for transforming ComponentGraph to ComponentDesc
    // ----------------------------------------------------------------------
    
    /**
     * Return the {@link ComponentDesc} corresponding to the specified ADL
     * descriptor.
     */
    private ComponentDesc toComponentDesc( String adl )
    throws Exception {
    	
    	JuliacConfig jconf = jc.getJuliacConfig();
    	String[] srcs = jconf.getSrcs();

    	/*
    	 * -e: dump stack trace
    	 * TODO Handle multiple source paths
    	 */
    	String[] args = new String[]{"-src-path="+srcs[0],adl,"-e"};
    	MindLauncher ml = new MindLauncher(args);
    	ComponentGraph cgraph = ml.getComponentGraph();
    	ComponentDesc cdesc = toComponentDesc(cgraph);
    	
        return cdesc;
    }
    
    private int compidx = 0;

	private ComponentDesc toComponentDesc( ComponentGraph cgraph )
	throws
		NoSuchFieldException, IllegalAccessException, InstantiationException,
		IOException, NoSuchInterfaceException {

		Node def = cgraph.getDefinition();

		String id = "C" + (compidx++);
		String name = getDeclaredValue(def, "_name");
		String definition = name;
		ComponentType ct = toType(def);
		String ctrlDesc = def.astGetType();
		String contentClassName = toContentDesc(def);

		ComponentDesc cdesc =
			new ComponentDesc(id,name,definition,ct,ctrlDesc,contentClassName);

		/*
		 * Subcomponents.
		 */
		Map<String,ComponentDesc> subcdescs =
			new HashMap<String,ComponentDesc>();
		ComponentGraph[] subcgraphs = cgraph.getSubComponents();
		for (ComponentGraph subcgraph : subcgraphs) {
			ComponentDesc subcdesc = toComponentDesc(subcgraph);
			String subname = subcgraph.getNameInParent(cgraph);
			subcdescs.put(subname,subcdesc);
			cdesc.addSubComponent(subcdesc);
		}

		/*
		 * Bindings.
		 */
		try {
			List<Object> _bindings = getDeclaredValue(def, "_bindings");

			for (Object _binding : _bindings) {

				String _fromComponentName =
					getDeclaredValue(_binding,"_fromComponent");
				String _toComponentName =
					getDeclaredValue(_binding,"_toComponent");
				String cltItfName = getDeclaredValue(_binding,"_fromInterface");
				String srvItfName = getDeclaredValue(_binding,"_toInterface");

				BindingType bindingType =
					toBindingType( _fromComponentName, _toComponentName );
				ComponentDesc clt =
					toComponentDesc(cdesc, subcdescs, _fromComponentName );
				ComponentDesc srv =
					toComponentDesc(cdesc, subcdescs, _toComponentName );

				BindingDesc bdesc = new BindingDesc(bindingType, clt,
						cltItfName, srv, srvItfName);
				clt.putBinding(cltItfName, bdesc);
			}
		} catch (NoSuchFieldException nsfe) {
			// No binding for this component
		}

		/*
		 * Attributes.
		 */
		try {
			List<Object> _attributes = getDeclaredValue(def,"_attributes");

			for (Object _attribute : _attributes) {
				String attrname = getDeclaredValue(_attribute, "_name");
				Object _value = getDeclaredValue(_attribute, "_value");
				
				/*
				 * Strip string delimiters which are included in the model built
				 * by the MIND ADL parser. Since this convention has not been
				 * taken by the Fractal ADL parser, the ComponentDesc model
				 * associated with Juliac assumes raw (undelimited) attribute
				 * values. Note that delimiters were not included in version
				 * 0.1-beta-3 of the MIND ADL toolchain. They seem to have
				 * appeared in version 0.1.
				 */
				String _v = getDeclaredValue(_value, "_value");
				if( _v.charAt(0)=='"' && _v.charAt(_v.length()-1)=='"' ) {
					_v = _v.substring(1,_v.length()-1);
				}
				
				InterfaceType it =
					ct.getFcInterfaceType("attribute-controller");
				String attributeController = it.getFcItfSignature();
				AttributeDesc adesc =
					new AttributeDesc(
						cdesc,attributeController,attrname,_v,null);
				cdesc.putAttribute(attrname,adesc);
			}
		} catch (NoSuchFieldException nsfe) {
			// No attribute for this component
		}

		return cdesc;
	}

	private BindingType toBindingType(
		String _fromComponentName, String _toComponentName )
	throws JuliacRuntimeException {

		if (_fromComponentName.equals("this")) {

			if (_toComponentName.equals("this")) {
				String msg = "Unsupported this to this binding";
				throw new JuliacRuntimeException(msg);
			}

			return BindingType.EXPORT;
		}

		if (_toComponentName.equals("this")) {
			return BindingType.IMPORT;
		}

		return BindingType.NORMAL;
	}

	private ComponentDesc toComponentDesc(
		ComponentDesc cdesc, Map<String, ComponentDesc> subcdescs,
		String name ) {
		
		if (name.equals("this")) {
			return cdesc;
		}
		ComponentDesc subcdesc = subcdescs.get(name);
		return subcdesc;
	}

	private static <T> T getDeclaredValue( Object o, String name )
	throws NoSuchFieldException, IllegalAccessException {
		Class<?> cl = o.getClass();
		Field field = cl.getDeclaredField(name);
		field.setAccessible(true);
		T value = (T) field.get(o);
		return value;
	}

	private static ComponentType toType( Node def )
	throws
		NoSuchFieldException, IllegalAccessException, InstantiationException {

		ComponentType type = null;

		List<?> _interfaces = getDeclaredValue(def, "_interfaces");
		int size = _interfaces.size();

		InterfaceType[] its = new InterfaceType[size];
		for (int i = 0; i < size; i++) {
			Object interf = _interfaces.get(i);

			String name = getDeclaredValue(interf, "_name");
			if (name.equals("attribute_controller")) {
				name = "attribute-controller";
			}
			String signature = getDeclaredValue(interf, "_signature");
			Object role = getDeclaredValue(interf, "_role");
			Object contingency = getDeclaredValue(interf, "_contingency");
			Object cardinality = getDeclaredValue(interf, "_cardinality");
			boolean isClient = role.equals("client");
			boolean isOptional = contingency != null;
			boolean isCollection = cardinality != null;

			its[i] =
				new BasicInterfaceType(
					name, signature, isClient, isOptional, isCollection );
		}

		type = new BasicComponentType(its);
		return type;
	}

	private static String toContentDesc( Node definition )
	throws IOException, NoSuchFieldException, IllegalAccessException {
		
		Object controllerDesc = definition.astGetType();
		String contentDesc = null;
		if (controllerDesc.equals("primitive")) {
			List<?> _sources = getDeclaredValue(definition, "_sources");
			if (_sources.size() != 1) {
				String msg =
					"Only 1 source file expected. Got: " + _sources.size();
				throw new IOException(msg);
			}
			Object source = _sources.get(0);
			String path = getDeclaredValue(source, "_path");
			// Remove heading / and trailing .java
			path = path.substring(1, path.length() - 5);
			contentDesc = path.replace('/', '.');
		}
		return contentDesc;
	}
}
